package compressTools;
import java.util.ArrayList;
import java.util.List;
/*
 * This source is part of CompressTools (c) 2012 by Bregalad
 * RLEINC v1 is (c) 2012 by Joel Yliluoma, http://iki.fi/bisqwit/
 * 
 * This is open source and free software.
 * 
 * This class is a coding algorithm and implements the Encoding interface
 */
public class RLEINC1 implements Encoding {

	public List<String> getLabels() {
		//This encoding doesn't require any extra header/footer data
		return Main.labels;
	}

	//Encode all blocks
	public List<List<Byte>> Encode(List<ArrayList<Byte>> source) {
		List<List<Byte>> encoded = new ArrayList<List<Byte>>();

		for(int block=0; block<source.size(); block++) {
			List<Byte> encoded_blk = EncodeBlock(source.get(block));
			encoded.add(encoded_blk);
		}
		return encoded;
	}

	//Decode all blocks
	public List<List<Byte>> Decode(List<List<Byte>> encoded) {
		List<List<Byte>> decoded = new ArrayList<List<Byte>>();

		for(int block=0; block<encoded.size(); block++) {
			List<Byte> decoded_blk = DecodeBlock(encoded.get(block));
			decoded.add(decoded_blk);
		}
		return decoded;
	}

	//Encode a single block
	private static List<Byte> EncodeBlock(List<Byte> source)
	{
		List<Byte> out_buf = new ArrayList<Byte>();

		int[] runs = new int[ source.size() ];
		int[] seqs = new int[ source.size() ];

		/* Find where are all the runs, and where are all the successive sequences */
		int b = source.size();

		for(int a=0; a<b; ) // Find runs of 1 byte
		{
			byte c = source.get(a);
			int runlen;
			for(runlen=1; a+runlen < b && c == source.get(a+runlen); ++runlen) {}
			if(runlen > 0x81) runlen = 0x81;
			if(runlen >= 3) runs[a] = runlen;
			a += runlen;
		}

		for(int a=0; a<b; ) // Find runs of successive sequence
		{
			int c = source.get(a) + 1;
			int seqlen;
			for(seqlen=1; a+seqlen<b && (c & 0xFF) == (source.get(a+seqlen) & 0xFF); ++seqlen) ++c;
			if(seqlen > 0x40) seqlen = 0x40;
			if(seqlen >= 3) seqs[a] = seqlen;
			a += seqlen;
		}

		/* Synthesize the result */
		for(int a=0; a<b; )
		{
			// Where is the next or the next sequence?
			int mb = a+0x100; if(mb>b) mb=b;
			int ra, rl=0; for(ra=a; ra<mb; ++ra) if((rl=runs[ra]) > 0) break;  if(rl<=0) ra=a+9999;
			int sa, sl=0; for(sa=a; sa<mb; ++sa) if((sl=seqs[sa]) > 0) break;  if(sl<=0) sa=a+9999;

			int time_until_run=ra-a; if(time_until_run<0) time_until_run=9999;
			int time_until_seq=sa-a; if(time_until_seq<0) time_until_seq=9999;
			int time_until = b-a;
			if(time_until_run < time_until) time_until = time_until_run;
			if(time_until_seq < time_until) time_until = time_until_seq;
			if(time_until >= 0x40) time_until = 0x40;

			if(time_until > 0)
			{
				// Do literal sequence. Store data in reverse.
				out_buf.add( (byte)(time_until-1) );
				for(int c=0; c<time_until; ++c)
					out_buf.add( source.get( a + time_until - (c+1) ) );
				a += time_until;
				continue;
			}

			// Do we have a sequence?
			if(time_until_seq == 0)
			{
				// Figure out its length
				int length = sl < 0x40 ? sl : 0x40;
				if(time_until_run < length) length = time_until_run<3 ? 3 : time_until_run;
				out_buf.add( (byte)( 0x40 + length - 1 ) );
				out_buf.add( source.get(a) );
				a += length;
			}
			else // We have a run
			{
				// Figure out its length
				int length = rl < 0x81 ? rl : 0x81;
				if(time_until_seq < length) length = time_until_seq<3 ? 3 : time_until_seq;
				out_buf.add( (byte)( 0x101 - length ) );
				out_buf.add( source.get(a) );
				a += length;
			}
		}
		// Append the end marker
		out_buf.add( (byte) 0x40 );
		return out_buf;
	}

	//Decode a single block
	private static List<Byte> DecodeBlock(List<Byte> encoded)
	{
		List<Byte> result = new ArrayList<Byte>();

		int end_where = -1;
		for(int i=0; i < encoded.size(); )
		{
			int c = encoded.get(i++) & 0xFF;
			if((c & 0x80) != 0) // RUN
			{
				for(int b = encoded.get(i++), n = 0x101-c; n > 0; --n)
					result.add( (byte)b );
			}
			else
			{
				if((c & 0x40) != 0) // SEQ
				{
					if(c == 0x40) { end_where = i; break; }
					for(int b = encoded.get(i++), n = c-0x3F; n > 0; --n)
						result.add( (byte)(b++ & 0xFF) );
				}
				else // LIT
				{
					for(int n = c; n >= 0; --n)
						result.add( encoded.get(i + n) );
					i += (c + 1);
				}
			}
		}
		if(end_where != encoded.size())
		{
			System.out.println("RLEINC1: An end tag was not the last item in the input data");
		}
		return result;
	}
}
